//========================================================================================
//
//	File:				FWClaInf.cpp
//	Release Version:	$ 1.0d11 $
//
//	Copyright:	(c) 1993, 1995 by Apple Computer, Inc., all rights reserved.
//
//========================================================================================

#include "FWFound.hpp"

#ifndef   FWCLAINF_H
#include "FWClaInf.h"
#endif

#ifndef   FWPRIDEB_H
#include "FWPriDeb.h"
#endif

#if FW_LIB_EXPORT_PRAGMAS
#pragma lib_export on
#endif

#ifdef FW_BUILD_MAC
#pragma segment FW_ClassInfo
#endif


//========================================================================================
// CLASS FW_CClassInfo
//========================================================================================

//----------------------------------------------------------------------------------------
//	FW_CClassInfo::FW_CClassInfo
//----------------------------------------------------------------------------------------
FW_CClassInfo::FW_CClassInfo(const char *const className,
							size_t instanceSize, 
							FW_ClassInfoPtr  const *ancestors,
							const size_t * const offsets) :
	fClassName(className),
	fInstanceSize(instanceSize),
	fAncestors(ancestors),
	fAncestorOffsets(offsets)
{
}

//----------------------------------------------------------------------------------------
//	FW_CClassInfo::~FW_CClassInfo
//----------------------------------------------------------------------------------------

FW_CClassInfo::~FW_CClassInfo()
{
}

//----------------------------------------------------------------------------------------
//	FW_CClassInfo::IsKindOf
//----------------------------------------------------------------------------------------

// (MLH) ??? - We should consider a cache scheme or a preprocessor calculation
// What we really want is a list of all the classes between this and
// the root. This knowledge is known at compile time. For the runtime
// cache case, instead of recursively calling this routine, call a routine
// that builds the list and stores it in a cache.
//
// The list is simply a vector of classinfo pointers and a count. It is much
// faster to scan one vector then to recursively call IsKindOf.
//
// (JEL) ??? - I don't agree with the above.  In order to flatten everything out
// into a vectors, we would have to duplicate a lot of data.  Suppose we have two
// classes X and Y, that each are derived from a deep hierarchy A -> B -> C.  To
// flatten into vectors, we would need two vectors [A,B,C,X] and [A,B,C,Y], duplicating
// the data [A,B,C].  I don't think it makes sense to tradeoff space for speed.
// In some specific uses it will make sense to cache, but I don't think that belongs
// here.  Perhaps we should take the effort to rewrite this algorithm so that it doesn't
// use this kind of direct recursion, by writing a loop that manipulates an internal
// stack that contains the state of the search, but I think that is overkill without
// evidence from testing that this function is a bottleneck.

FW_Boolean FW_CClassInfo::IsKindOf(FW_ClassInfoPtr aBaseClass) const
{
	FW_Boolean result = FALSE;
	
	if (aBaseClass == this)
		result = TRUE;
	else
	{
		FW_ClassInfoPtr const *ancestors = fAncestors;
	
		while (*ancestors != 0)
		{
			if ((*ancestors)->IsKindOf(aBaseClass))
			{
				result = TRUE;
				break;
			}
			ancestors++;
		}
	}
	return result;
}

FW_Boolean FW_CClassInfo::OffsetToBase(FW_ClassInfoPtr aBaseClass, size_t &offset) const
{
	FW_Boolean result = FALSE;
	
	if (aBaseClass == this)
	{
		offset = 0;
		result = TRUE;
	}
	else
	{
		FW_ClassInfoPtr const *ancestors = fAncestors;
	
		while (*ancestors != 0)
		{
			if ((*ancestors)->OffsetToBase(aBaseClass, offset))
			{
				offset += fAncestorOffsets[ancestors-fAncestors];
				result = TRUE;
				break;
			}
			ancestors++;
		}
	}
	
	return result;
}

//----------------------------------------------------------------------------------------
//	FW_CClassInfo::DynamicCast
//----------------------------------------------------------------------------------------

void* FW_CClassInfo::DynamicCast(const void* p, FW_ClassInfoPtr pointersClass, 
										FW_ClassInfoPtr aBaseClass) const
{
	size_t offset;
	char *q = 0;

	if ((pointersClass == aBaseClass) && IsKindOf(aBaseClass))
	{
		q = (char*) p;
	}
	else if (OffsetToBase(aBaseClass, offset))
	{
		q = (char*) p;
		q += offset;
		FW_Boolean isKindOf = OffsetToBase(pointersClass, offset);
		if (!isKindOf)
		{
			FW_PRIV_ASSERT(isKindOf);
				// If this assertion fails, the declared pointer type was 
				// not consistent with the runtime type of the object!
				// That could only happen if somewhere else a bad cast 
				// (perhaps from a void*) was done.
			return 0;
		}
		q -= offset;
	}

	return q;
}

// Consider this code fragment:
// 
// template <class TActual, class TKnown, class TTarget>
// foo()
// {
// 	TKnown *p = new TActual();
// 	TTarget *t = dynamic_cast<TTarget*>(p);
// }
// 
// TActual is the actual runtime type of the object.
// TKnown is the compile-time type of the pointer variable that points
//   to the object.
// TTarget is the compile-time type that we are attempting to cast
//   the pointer to.
// 
// Now, assume an operator IsKindOf(TDerived, TBase) that returns true
// if class TDerived is derived from class TBase.
// 
// Given this, the result of the dynamic_cast above will be nonzero
// if and only if IsKindOf(TActual, TTarget).  However, there are
// interesting subvariations. In the following, assume
// IsKindOf(TActual, TTarget):
// 
// The cast is a simple upcast if IsKindOf(TKnown, TTarget).  Upcasting
// never needs a dynamic_cast, i.e. it can be done with a simple cast
// (TTarget*).  Since ODF uses emulation, it is better to avoid
// dynamic casts in this instance.
// 
// The cast is a standard downcast if IsKindOf(TTarget, TKnown).  This
// is the most common case for dynamic_casts.
// 
// The cast is a sidecast if it is neither an upcast or a downcast, i.e.
// a valid sidecast requires three conditions:
// 
// 	IsKindOf(TActual, TTarget) && 
// 	!IsKindOf(TKnown, TTarget) &&
// 	!IsKindOf(TTarget, TKnown)
// 
// Sidecasts happen with mixin style programming.  Assume classes
// CBase, MMixin, and CDerived, where CDerived inherits from both
// CBase and MMixin.  A CBase* that points to a CDerived can be
// successfully cast to a MMixin*, even if the code that contains
// the dynamic_cast has no knowledge of the class CDerived, so
// there is apparently no relationship between the known type and
// the target type.
